"""Classes for the configuration of hermes-audio-server."""
import json
from pathlib import Path

from hermes_audio_server.config.mqtt import MQTTConfig
from hermes_audio_server.config.vad import VADConfig
from hermes_audio_server.exceptions import ConfigurationFileNotFoundError


# Default values
DEFAULT_CONFIG = '/etc/hermes-audio-server.json'
DEFAULT_SITE = 'default'

# Keys in the JSON configuration file
SITE = 'site'
MQTT = 'mqtt'
VAD = 'vad'


# TODO: Define __str__() with explicit settings for debugging.
class ServerConfig:
    """This class represents the configuration of a Hermes audio server.

    Attributes:
        site (str): The site ID of the audio server.
        mqtt (:class:`.MQTTConfig`): The MQTT options of the configuration.
        vad (:class:`.VADConfig`): The VAD options of the configuration.
    """

    def __init__(self, site='default', mqtt=None, vad=None):
        """Initialize a :class:`.ServerConfig` object.

        Args:
            site (str): The site ID of the Hermes audio server. Defaults
                to 'default'.
            mqtt (:class:`.MQTTConfig`, optional): The MQTT connection
                settings. Defaults to a default :class:`.MQTTConfig` object.
            vad (:class:`.VADConfig`, optional): The VAD settings. Defaults
                to a default :class:`.VADConfig` object, which disables voice
                activity detection.
        """
        if mqtt is None:
            self.mqtt = MQTTConfig()
        else:
            self.mqtt = mqtt

        if vad is None:
            self.vad = VADConfig()
        else:
            self.vad = vad

        self.site = site

    @classmethod
    def from_json_file(cls, filename=None):
        """Initialize a :class:`.ServerConfig` object with settings
        from a JSON file.

        Args:
            filename (str): The filename of a JSON file with the settings.
                Defaults to '/etc/hermes-audio-server'.

        Returns:
            :class:`.ServerConfig`: An object with the settings
            of the Hermes Audio Server.

        The :attr:`mqtt` attribute of the :class:`.ServerConfig`
        object is initialized with the MQTT connection settings from the
        configuration file, or the default values (hostname 'localhost' and
        port number 1883) if the settings are not specified.

        The :attr:`site` attribute of the :class:`.ServerConfig`
        object is initialized with the setting from the configuration file,
        or 'default' is the setting is not specified.

        The :attr:`vad` attribute of the :class:`.ServerConfig` object is
        initialized with the settings from the configuration file, or not
        enabled when not specified.

        Raises:
            :exc:`ConfigurationFileNotFoundError`: If :attr:`filename` doesn't
                exist.

            :exc:`PermissionError`: If we have no read permissions for
                :attr:`filename`.

            :exc:`JSONDecodeError`: If :attr:`filename` doesn't have a valid
                JSON syntax.

        The JSON file should have the following format:

        {
            "site": "default",
            "mqtt": {
                "host": "localhost",
                "port": 1883,
                "authentication": {
                    "username": "foobar",
                    "password": "secretpassword"
                },
                "tls": {
                    "ca_certificates": "",
                    "client_certificate": "",
                    "client_key": ""
                }
            },
            "vad": {
                "mode": 0,
                "silence": 2,
                "status_messages": true
            }
        }
        """
        if not filename:
            filename = DEFAULT_CONFIG

        try:
            with Path(filename).open('r') as json_file:
                configuration = json.load(json_file)
        except FileNotFoundError as error:
            raise ConfigurationFileNotFoundError(error.filename)

        return cls(site=configuration.get(SITE, DEFAULT_SITE),
                   mqtt=MQTTConfig.from_json(configuration.get(MQTT)),
                   vad=VADConfig.from_json(configuration.get(VAD)))



"""Module with the Hermes audio player class."""
import io
import json
import wave

from humanfriendly import format_size

from hermes_audio_server.exceptions import NoDefaultAudioDeviceError
from hermes_audio_server.mqtt import MQTTClient

PLAY_BYTES = 'hermes/audioServer/{}/playBytes/+'
PLAY_FINISHED = 'hermes/audioServer/{}/playFinished'
CHUNK = 256


class AudioPlayer(MQTTClient):
    """This class creates an MQTT client that acts as an audio player for the
    Hermes protocol.
    """

    def initialize(self):
        """Initialize a Hermes audio player."""
        self.logger.debug('Probing for available output devices...')
        for index in range(self.audio.get_device_count()):
            device = self.audio.get_device_info_by_index(index)
            name = device['name']
            channels = device['maxOutputChannels']
            if channels:
                self.logger.debug('[%d] %s', index, name)
        try:
            self.audio_out = self.audio.get_default_output_device_info()['name']
        except OSError:
            raise NoDefaultAudioDeviceError('output')
        self.logger.info('Connected to audio output %s.', self.audio_out)

    def on_connect(self, client, userdata, flags, result_code):
        """Callback that is called when the audio player connects to the MQTT
        broker."""
        super().on_connect(client, userdata, flags, result_code)
        # Listen to the MQTT topic defined in the Hermes protocol to play a WAV
        # file.
        # See https://docs.snips.ai/reference/hermes#playing-a-wav-sound
        play_bytes = PLAY_BYTES.format(self.config.site)
        self.mqtt.subscribe(play_bytes)
        self.mqtt.message_callback_add(play_bytes, self.on_play_bytes)
        self.logger.info('Subscribed to %s topic.', play_bytes)

    def on_play_bytes(self, client, userdata, message):
        """Callback that is called when the audio player receives a PLAY_BYTES
        message on MQTT.
        """
        request_id = message.topic.split('/')[4]
        length = format_size(len(message.payload), binary=True)
        self.logger.info('Received an audio message of length %s'
                         ' with request id %s on site %s.',
                         length,
                         request_id,
                         self.config.site)

        with io.BytesIO(message.payload) as wav_buffer:
            try:
                with wave.open(wav_buffer, 'rb') as wav:
                    sample_width = wav.getsampwidth()
                    sample_format = self.audio.get_format_from_width(sample_width)
                    n_channels = wav.getnchannels()
                    frame_rate = wav.getframerate()

                    self.logger.debug('Sample width: %s', sample_width)
                    self.logger.debug('Channels: %s', n_channels)
                    self.logger.debug('Frame rate: %s', frame_rate)

                    self.logger.debug('Opening audio output stream...')
                    stream = self.audio.open(format=sample_format,
                                             channels=n_channels,
                                             rate=frame_rate,
                                             output=True)

                    self.logger.debug('Playing WAV buffer on audio output...')
                    data = wav.readframes(CHUNK)

                    while data:
                        stream.write(data)
                        data = wav.readframes(CHUNK)

                    stream.stop_stream()
                    self.logger.debug('Closing audio output stream...')
                    stream.close()

                    self.logger.info('Finished playing audio message with id %s'
                                     ' on device %s on site %s.',
                                     request_id,
                                     self.audio_out,
                                     self.config.site)

                    # Publish a message that the audio service has finished
                    # playing the sound.
                    # See https://docs.snips.ai/reference/hermes#being-notified-when-sound-has-finished-playing
                    # This implementation doesn't publish a session ID.
                    play_finished_topic = PLAY_FINISHED.format(self.config.site)
                    play_finished_message = json.dumps({'id': request_id,
                                                        'siteId': self.config.site})
                    self.mqtt.publish(play_finished_topic,
                                      play_finished_message)
                    self.logger.debug('Published message on MQTT topic:')
                    self.logger.debug('Topic: %s', play_finished_topic)
                    self.logger.debug('Message: %s', play_finished_message)
            except wave.Error as error:
                self.logger.warning('%s', str(error))
            except EOFError:
                self.logger.warning('End of WAV buffer')



"""This module specifies some information about the package."""
PROJECT = 'hermes-audio-server'
DESCRIPTION = 'An open source implementation of the audio server part of the Hermes protocol'
KEYWORDS = 'hermes snips python3 rhasspy audio-player audio-recorder audio-server snips-audio-server hermes-protocol voice voice-control'
AUTHOR = 'Koen Vervloesem'
EMAIL = 'koen@vervloesem.eu'
LICENSE = 'MIT License'
GITHUB_URL = 'https://github.com/koenvervloesem/hermes-audio-server'
DOC_URL = 'https://github.com/koenvervloesem/hermes-audio-server'
TRACKER_URL = 'https://github.com/koenvervloesem/hermes-audio-server/issues'
PLAYER = 'hermes-audio-player'
RECORDER = 'hermes-audio-recorder'
VERSION = '0.2.0'



"""This module contains exceptions defined for Hermes Audio Server."""


class HermesAudioServerError(Exception):
    """Base class for exceptions raised by Hermes Audio Server code.

    By catching this exception type, you catch all exceptions that are
    defined by the Hermes Audio Server code."""


class ConfigurationFileNotFoundError(HermesAudioServerError):
    """Raised when the configuration file is not found."""

    def __init__(self, filename):
        """Initialize the exception with a string representing the filename."""
        self.filename = filename


class NoDefaultAudioDeviceError(HermesAudioServerError):
    """Raised when there's no default audio device available."""

    def __init__(self, inout):
        """Initialize the exception with a string representing input or output.
        """
        self.inout = inout


class UnsupportedPlatformError(HermesAudioServerError):
    """Raised when the platform Hermes Audio Server is running on is not
    supported."""

    def __init__(self, platform):
        """Initialize the exception with a string representing the platform."""
        self.platform = platform



"""Module with an MQTT client. Both the audio player and audio recorder class
inherit from this class.
"""
from paho.mqtt.client import Client
import pyaudio


class MQTTClient:
    """This class represents an MQTT client for Hermes Audio Server.

    This is an abstract base class. You don't instantiate an object of this
    class, but an object of one of its subclasses.
    """

    def __init__(self, config, verbose, logger):
        """Initialize an MQTT client.

        Args:
            config (:class:`.ServerConfig`): The configuration of
                the MQTT client.
            verbose (bool): Whether or not the MQTT client runs in verbose
                mode.
            logger (:class:`logging.Logger`): The Logger object for logging
                messages.
        """
        self.config = config
        self.verbose = verbose
        self.logger = logger
        self.mqtt = Client()
        self.logger.debug('Using %s', pyaudio.get_portaudio_version_text())
        self.logger.debug('Creating PyAudio object...')
        self.audio = pyaudio.PyAudio()

        self.initialize()

        self.mqtt.on_connect = self.on_connect
        self.mqtt.on_disconnect = self.on_disconnect
        self.connect()

    def connect(self):
        """Connect to the MQTT broker defined in the configuration."""
        # Set up MQTT authentication.
        if self.config.mqtt.auth.enabled:
            self.logger.debug('Setting username and password for MQTT broker.')
            self.mqtt.username_pw_set(self.config.mqtt.auth.username,
                                      self.config.mqtt.auth.password)

        # Set up an MQTT TLS connection.
        if self.config.mqtt.tls.enabled:
            self.logger.debug('Setting TLS connection settings for MQTT broker.')
            self.mqtt.tls_set(ca_certs=self.config.mqtt.tls.ca_certs,
                              certfile=self.config.mqtt.tls.client_cert,
                              keyfile=self.config.mqtt.tls.client_key)

        self.logger.debug('Connecting to MQTT broker %s:%s...',
                          self.config.mqtt.host,
                          self.config.mqtt.port)
        self.mqtt.connect(self.config.mqtt.host, self.config.mqtt.port)

    def initialize(self):
        """Initialize the MQTT client."""

    def start(self):
        """Start the event loop to the MQTT broker so the audio server starts
        listening to MQTT topics and the callback methods are called.
        """
        self.logger.debug('Starting MQTT event loop...')
        self.mqtt.loop_forever()

    def stop(self):
        """Disconnect from the MQTT broker and terminate the audio connection.
        """
        self.logger.debug('Disconnecting from MQTT broker...')
        self.mqtt.disconnect()
        self.logger.debug('Terminating PyAudio object...')
        self.audio.terminate()

    def on_connect(self, client, userdata, flags, result_code):
        """Callback that is called when the client connects to the MQTT broker.
        """
        self.logger.info('Connected to MQTT broker %s:%s'
                         ' with result code %s.',
                         self.config.mqtt.host,
                         self.config.mqtt.port,
                         result_code)

    def on_disconnect(self, client, userdata, flags, result_code):
        """Callback that is called when the client connects from the MQTT
        broker."""
        # This callback doesn't seem to be called.
        self.logger.info('Disconnected with result code %s.', result_code)



"""Class for the VAD configuration of hermes-audio-server."""

# Default values
DEFAULT_MODE = 0
DEFAULT_SILENCE = 2
DEFAULT_STATUS_MESSAGES = False

# Keys in the JSON configuration file
MODE = 'mode'
SILENCE = 'silence'
STATUS_MESSAGES = 'status_messages'


# TODO: Define __str__() for each class with explicit settings for debugging.
class VADConfig:
    """This class represents the VAD settings for Hermes Audio Recorder.

    Attributes:
        enabled (bool): Whether or not VAD is enabled.
        mode (int): Aggressiveness mode for VAD. 0 is the least aggressive
            about filtering out non-speech, 3 is the most aggressive.
        silence (int): How much silence (no speech detected) in seconds has
            to go by before Hermes Audio Recorder considers it the end of a
            voice message.
        status_messages (bool): Whether or not Hermes Audio Recorder sends
            messages on MQTT when it detects the start or end of a voice
            message.
    """

    def __init__(self, enabled=False, mode=0, silence=2, status_messages=False):
        """Initialize a :class:`.VADConfig` object.

        Args:
            enabled (bool): Whether or not VAD is enabled. Defaults to False.
            mode (int): Aggressiveness mode for VAD. Defaults to 0.
            silence (int): How much silence (no speech detected) in seconds has
                to go by before Hermes Audio Recorder considers it the end of a
                voice message. Defaults to 2.
            status_messages (bool): Whether or not Hermes Audio Recorder sends
                messages on MQTT when it detects the start or end of a voice
                message. Defaults to False.

        All arguments are optional.
        """
        self.enabled = enabled
        self.mode = mode
        self.silence = silence
        self.status_messages = status_messages

    @classmethod
    def from_json(cls, json_object=None):
        """Initialize a :class:`.VADConfig` object with settings from a
        JSON object.

        Args:
            json_object (optional): The JSON object with the VAD settings.
                Defaults to {}.

        Returns:
            :class:`.VADConfig`: An object with the VAD settings.

        The JSON object should have the following format:

        {
            "mode": 0,
            "silence": 2,
            "status_messages": true
        }
        """
        if json_object is None:
            ret = cls(enabled=False)
        else:
            ret = cls(enabled=True,
                      mode=json_object.get(MODE, DEFAULT_MODE),
                      silence=json_object.get(SILENCE, DEFAULT_SILENCE),
                      status_messages=json_object.get(STATUS_MESSAGES,
                                                      DEFAULT_STATUS_MESSAGES))

        return ret



"""Classes for the configuration of hermes-audio-server."""

# Default values
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 1883

# Keys in the JSON configuration file
HOST = 'host'
PORT = 'port'
AUTH = 'authentication'
USERNAME = 'username'
PASSWORD = 'password'
TLS = 'tls'
CA_CERTS = 'ca_certificates'
CLIENT_CERT = 'client_certificate'
CLIENT_KEY = 'client_key'


# TODO: Define __str__() for each class with explicit settings for debugging.
class MQTTAuthConfig:
    """This class represents the authentication settings for a connection to an
    MQTT broker.

    Attributes:
        username (str): The username to authenticate to the MQTT broker. `None`
            if there's no authentication.
        password (str): The password to authenticate to the MQTT broker. Can be
            `None`.
    """

    def __init__(self, username=None, password=None):
        """Initialize a :class:`.MQTTAuthConfig` object.

        Args:
            username (str, optional): The username to authenticate to the MQTT
                broker. `None` if there's no authentication.
            password (str, optional): The password to authenticate to the MQTT
                broker. Can be `None`.

        All arguments are optional.
        """
        self.username = username
        self.password = password

    @classmethod
    def from_json(cls, json_object=None):
        """Initialize a :class:`.MQTTAuthConfig` object with settings from a
        JSON object.

        Args:
            json_object (optional): The JSON object with the MQTT
                authentication settings. Defaults to {}.

        Returns:
            :class:`.MQTTAuthConfig`: An object with the MQTT authentication
            settings.

        The JSON object should have the following format:

        {
            "username": "foobar",
            "password": "secretpassword"
        }
        """
        if json_object is None:
            json_object = {}

        return cls(username=json_object.get(USERNAME),
                   password=json_object.get(PASSWORD))

    @property
    def enabled(self):
        """Check whether authentication is enabled.

        Returns:
            bool: True if the username is not `None`.
        """
        return self.username is not None


class MQTTTLSConfig:
    """This class represents the TLS settings for a connection to an MQTT
    broker.

    Attributes:
        enabled (bool): Whether or not TLS is enabled.
        ca_certs (str): Path to the Certificate Authority file. If `None`, the
            default certification authority of the system is used.
        client_key (str): Path to an PEM encoded private key file. If `None`,
            there will be no client authentication.
        client_cert (str): Path to a PEM encoded client certificate file. If
            `None`, there will be no client authentication.
    """

    def __init__(self, enabled=False, ca_certs=None, client_key=None,
                 client_cert=None):
        """Initialize a :class:`.MQTTTLSConfig` object.

        Args:
            enabled (bool, optional): Whether or not TLS is enabled. The
                default value is `False`.
            ca_certs (str, optional): Path to the Certificate Authority file.
                If `None`, the default certification authority of the system is
                used.
            client_key (str, optional): Path to a PEM encoded private key file.
                If `None`, there will be no client authentication.
            client_cert (str, optional): Path to a PEM encoded client
                certificate file. If `None`, there will be no client
                authentication.

        All arguments are optional.
        """
        self.enabled = enabled
        self.ca_certs = ca_certs
        self.client_key = client_key
        self.client_cert = client_cert

    @classmethod
    def from_json(cls, json_object=None):
        """Initialize a :class:`.MQTTTLSConfig` object with settings from a
        JSON object.

        Args:
            json_object (optional): The JSON object with the MQTT TLS settings.
                Defaults to {}.

        Returns:
            :class:`.MQTTTLSConfig`: An object with the MQTT TLS settings.

        The JSON object should have the following format:

        {
            "ca_certificates": "",
            "client_certificate": "",
            "client_key": ""
        }
        """
        if json_object is None:
            ret = cls(enabled=False)
        else:
            ret = cls(enabled=True,
                      ca_certs=json_object.get(CA_CERTS),
                      client_key=json_object.get(CLIENT_KEY),
                      client_cert=json_object.get(CLIENT_CERT))

        return ret

class MQTTConfig:
    """This class represents the configuration for a connection to an
    MQTT broker.

    Attributes:
        host (str): The hostname or IP address of the MQTT broker.
        port (int): The port number  of the MQTT broker.
        auth (:class:`.MQTTAuthConfig`, optional): The authentication
            settings (username and password) for the MQTT broker.
        tls (:class:`.MQTTTLSConfig`, optional): The TLS settings for the MQTT
            broker.
    """

    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT, auth=None,
                 tls=None):
        """Initialize a :class:`.MQTTConfig` object.

        Args:
            host (str, optional): The hostname or IP address of the MQTT
                broker.
            port (int, optional): The port number of the MQTT broker.
            auth (:class:`.MQTTAuthConfig`, optional): The authentication
                settings (username and password) for the MQTT broker. Defaults
                to a default :class:`.MQTTAuthConfig` object.
            tls (:class:`.MQTTTLSConfig`, optional): The TLS settings for the
                MQTT broker. Defaults to a default :class:`.MQTTTLSConfig`
                object.

        All arguments are optional.
        """
        self.host = host
        self.port = port

        if auth is None:
            self.auth = MQTTAuthConfig()
        else:
            self.auth = auth

        if tls is None:
            self.tls = MQTTTLSConfig()
        else:
            self.tls = tls

    @classmethod
    def from_json(cls, json_object=None):
        """Initialize a :class:`.MQTTConfig` object with settings from a JSON
        object.

        Args:
            json_object (optional): The JSON object with the MQTT settings.
                Defaults to {}.

        Returns:
            :class:`.MQTTConfig`: An object with the MQTT settings.

        The JSON object should have the following format:

        {
            "host": "localhost",
            "port": 1883,
            "authentication": {
                "username": "foobar",
                "password": "secretpassword"
            },
            "tls": {
                "ca_certificates": "",
                "client_certificate": "",
                "client_key": ""
            }
        }
        """
        if json_object is None:
            json_object = {}

        return cls(host=json_object.get(HOST, DEFAULT_HOST),
                   port=json_object.get(PORT, DEFAULT_PORT),
                   auth=MQTTAuthConfig.from_json(json_object.get(AUTH)),
                   tls=MQTTTLSConfig.from_json(json_object.get(TLS)))



"""This module contains helper functions to log messages from Hermes Audio
Server."""
import logging
from logging.handlers import SysLogHandler
import sys

import colorlog

from hermes_audio_server.exceptions import UnsupportedPlatformError

DAEMON_FORMAT = '{}[%(process)d]: %(message)s'
INTERACTIVE_FORMAT = '%(asctime)s %(log_color)s%(levelname)-8s%(reset)s %(message)s'
LOG_COLORS = {'DEBUG':    'white',
              'INFO':     'green',
              'WARNING':  'yellow',
              'ERROR':    'red',
              'CRITICAL': 'bold_red'}


def get_domain_socket():
    """Get the default domain socket for syslog on this platform."""
    if sys.platform.startswith('linux'):  # Linux
        return '/dev/log'
    if sys.platform.startswith('darwin'):  # macOS
        return '/var/run/syslog'
    # Unsupported platform
    raise UnsupportedPlatformError(sys.platform)


def get_logger(command, verbose, daemon):
    """Return a Logger object with the right level, formatter and handler."""

    if daemon:
        handler = SysLogHandler(address=get_domain_socket())
        formatter = logging.Formatter(fmt=DAEMON_FORMAT.format(command))
        logger = logging.getLogger(command)
    else:
        handler = colorlog.StreamHandler(stream=sys.stdout)
        formatter = colorlog.ColoredFormatter(INTERACTIVE_FORMAT,
                                              log_colors=LOG_COLORS)
        logger = colorlog.getLogger(command)

    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger



"""This module contains the main function run by the CLI commands
hermes-audio-player and hermes-audio-recorder.
"""
from json import JSONDecodeError
import signal
import sys

from daemon import DaemonContext

from hermes_audio_server.about import VERSION
from hermes_audio_server.config import ServerConfig, DEFAULT_CONFIG
from hermes_audio_server.exceptions import ConfigurationFileNotFoundError, \
    NoDefaultAudioDeviceError, UnsupportedPlatformError
from hermes_audio_server.logger import get_logger
from hermes_audio_server.player import AudioPlayer
from hermes_audio_server.recorder import AudioRecorder

SERVER = {'hermes-audio-player': AudioPlayer,
          'hermes-audio-recorder': AudioRecorder}


def main(command, verbose, version, config, daemon):
    """The main function run by the CLI command.

    Args:
        command (str): The command to run.
        verbose (bool): Use verbose output if True.
        version (bool): Print version information and exit if True.
        config (str): Configuration file.
        daemon (bool): Run as a daemon if True.
    """
    # Define signal handler to cleanly exit the program.
    def exit_process(signal_number, frame):
        # pylint: disable=no-member
        logger.info('Received %s signal. Exiting...',
                    signal.Signals(signal_number).name)
        server.stop()
        sys.exit(0)

    # Register signals.
    signal.signal(signal.SIGQUIT, exit_process)
    signal.signal(signal.SIGTERM, exit_process)

    try:

        logger = get_logger(command, verbose, daemon)
        logger.info('%s %s', command, VERSION)

        # Start the program as a daemon.
        if daemon:
            logger.debug('Starting daemon...')
            context = DaemonContext(files_preserve=[logger.handlers[0].socket])
            context.signal_map = {signal.SIGQUIT: exit_process,
                                  signal.SIGTERM: exit_process}
            context.open()

        if not version:
            if not config:
                logger.debug('Using default configuration file.')
                config = DEFAULT_CONFIG

            server_class = SERVER[command]
            logger.debug('Creating %s object...', server_class.__name__)
            server = server_class(ServerConfig.from_json_file(config),
                                  verbose,
                                  logger)

            server.start()
    except ConfigurationFileNotFoundError as error:
        logger.critical('Configuration file %s not found. Exiting...', error.filename)
        sys.exit(1)
    except JSONDecodeError as error:
        logger.critical('%s is not a valid JSON file. Parsing failed at line %s and column %s. Exiting...', config, error.lineno, error.colno)
        sys.exit(1)
    except KeyboardInterrupt:
        logger.info('Received SIGINT signal. Shutting down %s...', command)
        server.stop()
        sys.exit(0)
    except NoDefaultAudioDeviceError as error:
        logger.critical('No default audio %s device available. Exiting...',
                        error.inout)
        sys.exit(1)
    except PermissionError as error:
        logger.critical('Can\'t read file %s. Make sure you have read permissions. Exiting...', error.filename)
        sys.exit(1)
    except UnsupportedPlatformError as error:
        # Don't use logger because this exception is thrown while logging.
        print('Error: {} is not a supported platform.'.format(error.platform))
        sys.exit(1)
